
#include "stdafx.h"
#include "video.h"
#include "CodecCtl.h"

#include <vfw.h>

#define ERRORNR_CODEC_BADFORMAT 1201
#define ERRORSTR_CODEC_BADFORMAT "Format Not Supported"
#define ERRORNR_CODEC_LOCATE 1202
#define ERRORSTR_CODEC_LOCATE "Unable to Open the following Codec: "
#define ERRORNR_CODEC_COMPRESS 1203
#define ERRORSTR_CODEC_COMPRESS "Compress-Error"
#define ERRORNR_CODEC_DECOMPRESS 1204
#define ERRORSTR_CODEC_DECOMPRESS "Decompress-Error"
#define ERRORNR_CODEC_INVALID 1205
#define ERRORSTR_CODEC_INVALID "Invalid Parameter"

long bmpsize(BITMAPINFOHEADER* bmph)
{
   return (bmph->biCompression==BI_RGB)?bmph->biWidth*bmph->biHeight*bmph->biBitCount/8:bmph->biSizeImage;
}

long bmphsize(BITMAPINFOHEADER* bmph)
{
   return bmph->biSize+bmph->biClrUsed*sizeof(RGBQUAD);
}

unsigned long fourchar(char *s)
{
   union u { char uc[4]; unsigned long ul; } uu;
   for(int n=0;n<4;n++) uu.uc[n]=s[n]; 
   return uu.ul;
}

CString fourchar(unsigned long l)
{
   CString s;
   union u { char uc[5]; unsigned long ul; } uu;
   uu.ul=l; uu.uc[4]=0;
   s=uu.uc;
   return s;
}

IMPLEMENT_DYNCREATE(CCodecCtrl, COleControl)

BEGIN_MESSAGE_MAP(CCodecCtrl, COleControl)
	//{{AFX_MSG_MAP(CCodecCtrl)
	//}}AFX_MSG_MAP
	ON_OLEVERB(AFX_IDS_VERB_PROPERTIES, OnProperties)
END_MESSAGE_MAP()

BEGIN_DISPATCH_MAP(CCodecCtrl, COleControl)
	//{{AFX_DISPATCH_MAP(CCodecCtrl)
	DISP_PROPERTY(CCodecCtrl, "Quality", m_quality, VT_I2)
	DISP_PROPERTY_NOTIFY(CCodecCtrl, "Codec", m_codec, OnCodecChanged, VT_BSTR)
	DISP_PROPERTY(CCodecCtrl, "FrameNr", m_frameNr, VT_I2)
	DISP_PROPERTY_NOTIFY(CCodecCtrl, "KeyFrame", m_keyFrame, OnKeyFrameChanged, VT_I2)
	DISP_FUNCTION(CCodecCtrl, "EncodeFrame", EncodeFrame, VT_EMPTY, VTS_I4 VTS_I4 VTS_I4)
	DISP_FUNCTION(CCodecCtrl, "DecodeFrame", DecodeFrame, VT_EMPTY, VTS_I4 VTS_I4 VTS_I4)
	DISP_FUNCTION(CCodecCtrl, "SelectCodec", SelectCodec, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CCodecCtrl, "NoCodec", NoCodec, VT_EMPTY, VTS_NONE)
	//}}AFX_DISPATCH_MAP
	DISP_FUNCTION_ID(CCodecCtrl, "AboutBox", DISPID_ABOUTBOX, AboutBox, VT_EMPTY, VTS_NONE)
END_DISPATCH_MAP()

BEGIN_EVENT_MAP(CCodecCtrl, COleControl)
	//{{AFX_EVENT_MAP(CCodecCtrl)
	EVENT_CUSTOM("OnFrame", FireOnFrame, VTS_I4  VTS_I4  VTS_I4)
	EVENT_CUSTOM("Resize", FireResize, VTS_I4  VTS_I4)
	//}}AFX_EVENT_MAP
END_EVENT_MAP()

IMPLEMENT_OLECREATE_EX(CCodecCtrl, "VIDEO.CodecCtrl.1", 0x96872fb3, 0xd44c, 0x11d1, 0xa8, 0x98, 0, 0x20, 0xaf, 0xf8, 0x4c, 0xb3)
IMPLEMENT_OLETYPELIB(CCodecCtrl, _tlid, _wVerMajor, _wVerMinor)

const IID BASED_CODE IID_DCodec = { 0x96872fb1, 0xd44c, 0x11d1, { 0xa8, 0x98, 0, 0x20, 0xaf, 0xf8, 0x4c, 0xb3 } };
const IID BASED_CODE IID_DCodecEvents = { 0x96872fb2, 0xd44c, 0x11d1, { 0xa8, 0x98, 0, 0x20, 0xaf, 0xf8, 0x4c, 0xb3 } };

static const DWORD BASED_CODE _dwCodecOleMisc =
	OLEMISC_INVISIBLEATRUNTIME |
	OLEMISC_SETCLIENTSITEFIRST |
	OLEMISC_INSIDEOUT |
	OLEMISC_CANTLINKINSIDE |
	OLEMISC_RECOMPOSEONRESIZE;

IMPLEMENT_OLECTLTYPE(CCodecCtrl, IDS_CODEC, _dwCodecOleMisc)

BOOL CCodecCtrl::CCodecCtrlFactory::UpdateRegistry(BOOL bRegister)
{
	if (bRegister)
		return AfxOleRegisterControlClass(
			AfxGetInstanceHandle(),
			m_clsid,
			m_lpszProgID,
			IDS_CODEC,
			IDB_CODEC,
			afxRegApartmentThreading,
			_dwCodecOleMisc,
			_tlid,
			_wVerMajor,
			_wVerMinor);
	else
		return AfxOleUnregisterClass(m_clsid, m_lpszProgID);
}

CCodecCtrl::CCodecCtrl()
{
   m_puffer=0; m_prevpuffer=0; m_bmph=0;
   m_begin=0; 
	InitializeIIDs(&IID_DCodec, &IID_DCodecEvents);
   OnResetState();
   memset(&m_compvars, 0, sizeof(COMPVARS));
   m_prevbmph=0;
}

CCodecCtrl::~CCodecCtrl()
{
   OnResetState();
   delete m_prevbmph;
}

void CCodecCtrl::OnDraw(CDC* pdc, const CRect& rcBounds, const CRect& rcInvalid)
{
   DrawOcxBitmap(this, pdc, rcBounds, IDB_CODEC);
}

void CCodecCtrl::DoPropExchange(CPropExchange* pPX)
{
	ExchangeVersion(pPX, MAKELONG(_wVerMinor, _wVerMajor));
	COleControl::DoPropExchange(pPX);
}

void CCodecCtrl::OnResetState()
{
	COleControl::OnResetState();  // Resets defaults found in DoPropExchange

   delete m_bmph;
   delete m_puffer;
   delete m_prevpuffer;
   m_puffer=0; m_prevpuffer=0; m_bmph=0;

   if (m_begin==1) ICCompressEnd(m_hic);
   if (m_begin==2) ICDecompressEnd(m_hic);
   if (m_hic) ICClose(m_hic);
   m_begin=0;
   m_hic=0;

   m_biCompression=0; m_Width=0; m_Height=0; m_BitCount=0;
   m_flag=0;
   m_prevpuffersize=0;
   m_frameNr=0; m_keyFrame=1; m_quality=7500;
}

void CCodecCtrl::AboutBox()
{
	CDialog dlgAbout(IDD_ABOUTBOX);
	dlgAbout.DoModal();
}

// ******************************************************************

void CCodecCtrl::EncodeFrame(long lbmph, long data, long size) 
{
   if (m_hic==0)
   {
      if (m_begin) { ICCompressEnd(m_hic); m_begin=0; }
      FireOnFrame(lbmph, data, size);
   }
   else
   {
      BITMAPINFOHEADER *bmph=(BITMAPINFOHEADER*)lbmph;

      if (m_biCompression!=bmph->biCompression || m_Width!=bmph->biWidth || m_Height!=bmph->biHeight || m_BitCount!=bmph->biBitCount || m_flag==1)
      {
         if (m_begin) { ICCompressEnd(m_hic); m_begin=0; }
         if (ICCompressQuery(m_hic, bmph, 0)!=ICERR_OK) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_BADFORMAT), ERRORSTR_CODEC_BADFORMAT, 0); return; }

         // Ausgangs-BmpH setzen
         long size=ICCompressGetFormatSize(m_hic, bmph);
         delete m_bmph;
         m_bmph=(BITMAPINFOHEADER*) new char[size];
         memset(m_bmph, 0, size);
         if (ICCompressGetFormat(m_hic, bmph, m_bmph)!=ICERR_OK) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_BADFORMAT), ERRORSTR_CODEC_BADFORMAT, 0); return; }

         // max. Gre des Puffers berechnen
         delete m_puffer;
         m_puffer=new char[ICCompressGetSize(m_hic, bmph, m_bmph)];

         m_biCompression=bmph->biCompression;
         m_Width=bmph->biWidth;
         m_Height=bmph->biHeight;
         m_BitCount=bmph->biBitCount;
         m_flag=0;
	      m_frameNr=0;
      }
      else
      {
         m_frameNr=(m_frameNr+1)%m_keyFrame;
      }
      
      if (!m_begin)
      {
         if (ICCompressBegin(m_hic, bmph, m_bmph)!=ICERR_OK) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_COMPRESS), ERRORSTR_CODEC_COMPRESS, 0); return; }
         m_begin=1;
      }

      if (!m_prevpuffersize) m_frameNr=0;
      DWORD Flags=(!m_frameNr)?ICCOMPRESS_KEYFRAME:0;
      DWORD FrameSize=ICCompressGetFormatSize(m_hic, bmph); //((m_icinfo.dwFlags&VIDCF_CRUNCH)?0:0);
      DWORD Quality=(m_icinfo.dwFlags&VIDCF_QUALITY)?m_quality:0;
      BITMAPINFO* bmphprev=(!m_frameNr)?0:(BITMAPINFO*)m_prevbmph;
      void* dataprev=(!m_frameNr)?0:m_prevpuffer;
    
      if (!m_prevpuffersize || !(m_icinfo.dwFlags&VIDCF_TEMPORAL) || (m_icinfo.dwFlags&VIDCF_FASTTEMPORALC) || m_keyFrame==1)
      {
         bmphprev=0; dataprev=0;
      }
         
      DWORD dwCkID, dwCompFlags; 
      if (ICCompress(m_hic, Flags, m_bmph, m_puffer, bmph, (void*)data, &dwCkID, &dwCompFlags, m_frameNr, FrameSize, Quality, (BITMAPINFOHEADER*)bmphprev, dataprev)!=ICERR_OK) 
         ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_COMPRESS), ERRORSTR_CODEC_COMPRESS, 0);
      else
         FireOnFrame((long)m_bmph, (long)m_puffer, bmpsize(m_bmph));

      //if (dwCompFlags & AVIIF_KEYFRAME) ...das war ein Key-Frame

      if ((m_icinfo.dwFlags&VIDCF_TEMPORAL) && !(VIDCF_FASTTEMPORALC&m_icinfo.dwFlags) && m_keyFrame>1)
      {
         //Copy Frame to PrevFrame
         if (size>m_prevpuffersize) { delete m_prevpuffer; m_prevpuffer=new char[m_prevpuffersize=size]; }
         memcpy(m_prevpuffer, (void*)data, size);
      
         delete m_prevbmph; m_prevbmph=new char[bmphsize(bmph)];
         memcpy(m_prevbmph, bmph, bmphsize(bmph));
      }
      else
      {
         delete m_prevbmph; m_prevbmph=0;
      }
   }
}

void CCodecCtrl::DecodeFrame(long lbmph, long data, long size) 
{
   BITMAPINFOHEADER *bmph=(BITMAPINFOHEADER*)lbmph;

   if (m_Width!=bmph->biWidth || m_Height!=bmph->biHeight) FireResize(bmph->biWidth, bmph->biHeight);

   if (bmph->biCompression==BI_RGB)
   {
      if (m_begin) { ICDecompressEnd(m_hic); m_begin=0; }
      FireOnFrame(lbmph, data, size);
      m_codec="";
   }
   else
   {
      if (m_biCompression!=bmph->biCompression || m_Width!=bmph->biWidth || m_Height!=bmph->biHeight || m_BitCount!=bmph->biBitCount || (long)m_hic==-1)
      {
         if (m_begin) { ICDecompressEnd(m_hic); m_begin=0; }

         // neuer Dekompressor
         if (m_hic) ICClose(m_hic);
         m_hic=ICLocate(ICTYPE_VIDEO, 0, bmph, 0, ICMODE_DECOMPRESS); if (!m_hic) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_LOCATE), ERRORSTR_CODEC_LOCATE+fourchar(bmph->biCompression), 0); return; }
         m_codec=fourchar(bmph->biCompression);
         if (ICDecompressQuery(m_hic, bmph, 0)!=ICERR_OK) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_DECOMPRESS), ERRORSTR_CODEC_DECOMPRESS, 0); ICClose(m_hic); m_hic=0; return; }

         // Ausgangs-BmpH setzen
         long size=ICDecompressGetFormat(m_hic, bmph, 0);
         delete m_bmph;
         m_bmph=(BITMAPINFOHEADER*) new char[size];
         memset(m_bmph, 0, size);
         if (ICDecompressGetFormat(m_hic, bmph, m_bmph)!=ICERR_OK) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_BADFORMAT), ERRORSTR_CODEC_BADFORMAT, 0); ICClose(m_hic); m_hic=0; return; }
         
         // Gre des Puffers berechnen
         delete m_puffer;
         m_puffer=new char[bmpsize(m_bmph)];
      }
      
      if (!m_begin)
      {
         if (ICDecompressBegin(m_hic, bmph, m_bmph)!=ICERR_OK) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_DECOMPRESS), ERRORSTR_CODEC_DECOMPRESS, 0); return; }
         m_begin=2;
      }

      if (ICDecompress(m_hic, 0, bmph, (void*)data, m_bmph, m_puffer)!=ICERR_OK) 
         ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_DECOMPRESS), ERRORSTR_CODEC_DECOMPRESS, 0);
      else
         FireOnFrame((long)m_bmph, (long)m_puffer, bmpsize(m_bmph));
   }  

   m_biCompression=bmph->biCompression;
   m_Width=bmph->biWidth;
   m_Height=bmph->biHeight;
   m_BitCount=bmph->biBitCount;
}

void CCodecCtrl::SelectCodec() 
{
   COMPVARS c; 
   c=m_compvars;
   c.cbSize=sizeof(COMPVARS);
   if (m_compvars.dwFlags & ICMF_COMPVARS_VALID)
   {
	   c.lKey=m_keyFrame;
      c.fccHandler=fourchar((char*)(LPCTSTR)m_codec);
      c.lQ=m_quality>=0?m_quality:0;
   }
   
   if (!ICCompressorChoose(m_hWnd, ICMF_CHOOSE_ALLCOMPRESSORS | ICMF_CHOOSE_DATARATE | ICMF_CHOOSE_KEYFRAME /*| ICMF_CHOOSE_PREVIEW*/, 0, 0, &c, "Select Codec")) return;

   NoCodec();
   m_compvars=c;
   m_hic=m_compvars.hic;
	m_frameNr=0;
	m_keyFrame=(short)m_compvars.lKey;
   m_flag=1;

   m_codec=fourchar(m_compvars.fccHandler);

   ICGetInfo(m_hic, &m_icinfo, sizeof(ICINFO));
   m_quality=(m_icinfo.dwFlags&VIDCF_QUALITY)?m_compvars.lQ:-1;
}

void CCodecCtrl::OnCodecChanged() 
{
   if (m_codec.GetLength()==0) { NoCodec(); return; }
   if (m_codec.GetLength()!=4) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_INVALID), ERRORSTR_CODEC_INVALID, 0); m_codec=""; return; }

   HIC hic=ICOpen(ICTYPE_VIDEO, fourchar((char*)(LPCTSTR)m_codec), ICMODE_COMPRESS);
   if (!hic) { ThrowError(CUSTOM_CTL_SCODE(ERRORNR_CODEC_LOCATE), ERRORSTR_CODEC_LOCATE+m_codec, 0); NoCodec(); return; }
   
   resetcodec();
   
   m_hic=hic;

   ICGetInfo(m_hic, &m_icinfo, sizeof(ICINFO));
   m_flag=1;

	SetModifiedFlag();
}

void CCodecCtrl::NoCodec() 
{
   m_codec="";
   resetcodec();
}

void CCodecCtrl::resetcodec(void)
{
   if (m_begin) { ICCompressEnd(m_hic); m_begin=0; }
   if (m_hic) { ICClose(m_hic); m_hic=0; }
   if (m_compvars.dwFlags & ICMF_COMPVARS_VALID) ICCompressorFree(&m_compvars);
  
   memset(&m_compvars, 0, sizeof(COMPVARS));

   delete m_bmph;       m_bmph=0;
   delete m_puffer;     m_puffer=0; 
   delete m_prevpuffer; m_prevpuffer=new char[m_prevpuffersize=20000];

   m_biCompression=0; m_Width=0; m_Height=0; m_BitCount=0;
   m_flag=0;
	m_frameNr=0; m_keyFrame=1;
}

void CCodecCtrl::OnKeyFrameChanged() 
{
   m_frameNr=0;
   if (m_keyFrame<1) m_keyFrame=1;
	SetModifiedFlag();
}
